"""
Script to count distinct values in a HXL dataset.
David Megginson
October 2014

Counts all combinations of the tags specified on the command line. In
the command-line version, you may omit the initial '#' from tag names
to avoid the need to quote them.  Also optionally calculates sum,
average (mean), min, and max for a numeric tag.

Only the *first* column with each hashtag is currently used.

License: Public Domain
Documentation: http://hxlstandard.org
"""

import sys
import argparse
from hxl.model import HXLDataProvider, HXLColumn, HXLRow
from hxl.filters import parse_tags, fix_tag, find_column
from hxl.parser import HXLReader, writeHXL

class HXLCountFilter(HXLDataProvider):
    """
    Composable filter class to aggregate rows in a HXL dataset.

    This is the class supporting the hxlcount command-line utility.

    Because this class is a {@link hxl.model.HXLDataProvider}, you can use
    it as the source to an instance of another filter class to build a
    dynamic, single-threaded processing pipeline.

    WARNING: this filter reads the entire source dataset before
    producing output, and may need to hold a large amount of data in
    memory, depending on the number of unique combinations counted.

    Usage:

    <pre>
    source = HXLReader(sys.stdin)
    filter = HXLCountFilter(source, tags=['#org', '#sector', '#adm1'])
    writeHXL(sys.stdout, filter)
    </pre>
    """

    def __init__(self, source, tags, aggregate_tag=None):
        """
        Constructor
        @param source the HXL data source
        @param tags a list of HXL tags that form a unique key together (what combinations are you counting?)
        @param aggregate_tag an optional numeric tag for calculating aggregate values.
        """
        self.source = source
        self.count_tags = tags
        self.aggregate_tag = aggregate_tag
        self.saved_columns = None
        self.aggregate_iter = None

    @property
    def columns(self):
        """
        @return the column definitions used in the aggregation report
        """
        if self.saved_columns is None:
            cols = []
            for tag in self.count_tags:
                column = find_column(tag, self.source.columns)
                if column:
                    headerText = column.headerText
                else:
                    headerText = None
                cols.append(HXLColumn(hxlTag=tag, headerText=headerText))
            cols.append(HXLColumn(hxlTag='#x_count_num', headerText='Count'))
            if self.aggregate_tag is not None:
                cols.append(HXLColumn(hxlTag='#x_sum_num', headerText='Sum'))
                cols.append(HXLColumn(hxlTag='#x_average_num', headerText='Average (mean)'))
                cols.append(HXLColumn(hxlTag='#x_min_num', headerText='Minimum value'))
                cols.append(HXLColumn(hxlTag='#x_max_num', headerText='Maximum value'))
            self.saved_columns = cols
        return self.saved_columns

    def __next__(self):
        """
        @return the next row of aggregated data.
        """
        if self.aggregate_iter is None:
            self._aggregate()
        # Write the stats, sorted in value order
        aggregate = next(self.aggregate_iter)
        values = list(aggregate[0])
        values.append(aggregate[1].count)
        if self.aggregate_tag:
            if aggregate[1].seen_numbers:
                values.append(aggregate[1].sum)
                values.append(aggregate[1].average)
                values.append(aggregate[1].min)
                values.append(aggregate[1].max)
            else:
                values = values + ([''] * 4)

        row = HXLRow(self.columns)
        row.values = values
        return row

    next = __next__

    def _aggregate(self):
        """
        Read the entire source dataset and produce saved aggregate data.
        """
        aggregators = {}
        for row in self.source:
            values = []
            for tag in self.count_tags:
                value = row.get(tag)
                if value is not False:
                    values.append(value)
            if values:
                key = tuple(values)
                if not key in aggregators:
                    aggregators[key] = Aggregator(self.aggregate_tag)
                aggregators[key].add(row)
        self.aggregate_iter = iter(sorted(aggregators.items()))

class Aggregator:
    """
    Class to collect aggregates for a single combination.

    Currently calculates count, sum, average, min, and max
    """

    def __init__(self, tag):
        """
        Constructor
        @param tag the HXL tag being counted in the row.
        """
        self.tag = tag
        self.count = 0
        self.sum = 0.0
        self.average = 0.0
        self.min = None
        self.max = None
        self.seen_numbers = False

    def add(self, row):
        """
        Add a new row of data to the aggregator.
        """
        self.count += 1
        if self.tag:
            value = row.get(self.tag)
            try:
                n = float(value)
                self.sum += n
                self.average = self.sum / self.count
                if self.min is None or n < self.min:
                    self.min = n
                if self.max is None or n > self.max:
                    self.max = n
                self.seen_numbers = True
            except ValueError:
                pass

#
# Command-line support
#

def run(args, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr):
    """
    Run hxlcount with command-line arguments.
    @param args A list of arguments, excluding the script name
    @param stdin Standard input for the script
    @param stdout Standard output for the script
    @param stderr Standard error for the script
    """

    # Command-line arguments
    parser = argparse.ArgumentParser(description = 'Generate aggregate counts for a HXL dataset')
    parser.add_argument(
        'infile',
        help='HXL file to read (if omitted, use standard input).',
        nargs='?',
        type=argparse.FileType('r'),
        default=stdin
        )
    parser.add_argument(
        'outfile',
        help='HXL file to write (if omitted, use standard output).',
        nargs='?',
        type=argparse.FileType('w'),
        default=stdout
        )
    parser.add_argument(
        '-t',
        '--tags',
        help='Comma-separated list of column tags to count.',
        metavar='tag,tag...',
        type=parse_tags,
        default='loc,org,sector,adm1,adm2,adm3'
        )
    parser.add_argument(
        '-a',
        '--aggregate',
        help='Hashtag to aggregate.',
        metavar='tag',
        type=fix_tag
        )
    args = parser.parse_args(args)
    source = HXLReader(args.infile)
    filter = HXLCountFilter(source, args.tags, args.aggregate)
    writeHXL(args.outfile, filter)

# end



"""
Parsing library for the Humanitarian Exchange Language (HXL) v1.0
David Megginson
October 2014

License: Public Domain
Documentation: http://hxlstandard.org
"""

import csv
import cgi
import json
import re
from .model import HXLDataProvider, HXLDataset, HXLColumn, HXLRow

class HXLParseException(Exception):
    """
    A parsing error in a HXL dataset.
    """

    def __init__(self, message, sourceRowNumber=None, sourceColumnNumber=None):
        super(Exception, self).__init__(message)
        self.sourceRowNumber = sourceRowNumber
        self.sourceColumnNumber = sourceColumnNumber

    def __str__(self):
        return '<HXLException: ' + str(self.message) + ' @ ' + str(self.sourceRowNumber) + ', ' + str(self.sourceColumnNumber) + '>'


class _TableSpec:
    """
    Table metadata for parsing a HXL dataset
    
    This class also contains logic for translating between raw source
    columns and logical output columns (taking into account untagged
    columns and compact-disaggregated syntax. It also caches
    calculated values for later reuse.
    """

    def __init__(self):
        self.colSpecs = []
        self.resetCache()

    def append(self, colSpec):
        """
        Append a new _ColSpec to the table spec
        """
        self.colSpecs.append(colSpec)
        self.resetCache()

    @property
    def columns(self):
        """
        Fairly-complex function to extract logical columns from virtual ones:

        - remove any column without a HXL hashtag
        - include the *first* compact disaggregated column, expanded to two
        - remove any other compact disaggregated columns

        Use self.cachedColumns to save the result, so that it has to run only
        once, no matter how many times it's called.
        """
        if self.cachedColumns == None:
            self.cachedColumns = []
            seenFixed = False
            for colSpec in self.colSpecs:
                if not colSpec.column.hxlTag or (colSpec.fixedColumn and seenFixed):
                    continue
                if colSpec.fixedColumn:
                    self.cachedColumns.append(colSpec.fixedColumn)
                    seenFixed = True
                self.cachedColumns.append(colSpec.column)
        return self.cachedColumns

    @property
    def headers(self):
        """
        Get a simple list of header strings from the columns.
        """
        if self.cachedHeaders == None:
            self.cachedHeaders = list(map(lambda column: column.headerText, self.columns))
        return self.cachedHeaders

    @property
    def tags(self):
        """
        Get a simple list of HXL hashtags from the columns.
        """
        if self.cachedTags == None:
            self.cachedTags = list(map(lambda column: column.hxlTag, self.columns))
        return self.cachedTags

    @property
    def disaggregationCount(self):
        """
        Get the number of columns using compact disaggregated syntax
        
        They will all be merged into one logical column, with repeated rows
        for different values. Cache the result in self.cachedDisaggregationCount.
        """
        if self.cachedDisaggregationCount == None:
            self.cachedDisaggregationCount = 0;
            for colSpec in self.colSpecs:
                if colSpec.fixedColumn:
                    self.cachedDisaggregationCount += 1
        return self.cachedDisaggregationCount

    def getDisaggregatedSourceColumnNumber(self, disaggregationPosition):
        """
        For a logical sequence number in disaggregation, figure out the
        actual source column number.
        """
        for pos, colSpec in enumerate(self.colSpecs):
            if colSpec.fixedColumn:
                disaggregationPosition -= 1
            if disaggregationPosition < 0:
                return pos
        return -1

    def resetCache(self):
        """
        Reset all cached values so that they'll be regenerated.
        """
        self.cachedColumns = None
        self.cachedHeaders = None
        self.cachedTags = None
        self.cachedDisaggregationCount = None

    def __str__(self):
        s = '<_TableSpec';
        for colSpec in self.colSpecs:
            s += "\n  " + re.sub("\n", "\n  ", str(colSpec));
        s += "\n>"
        return s


class _ColSpec:
    """
    Column metadata for parsing a HXL CSV file

    This class captures the way a column is encoded in the input CSV
    file, which might be different from the logical structure of the
    HXL data. Used only during parsing.
    """

    def __init__(self, sourceColumnNumber, column=None, fixedColumn=None, fixedValue=None):
        self.sourceColumnNumber = sourceColumnNumber
        self.column = column
        self.fixedColumn = fixedColumn
        self.fixedValue = fixedValue

    def __str__(self):
        s = "<_ColSpec";
        s += "\n  column: " + str(self.column)
        if (self.fixedColumn):
            s += "\n  fixedColumn: " + str(self.fixedColumn)
            s += "\n  fixedValue: " + str(self.fixedValue)
        s += "\n  sourceColumnNumberumber: " + str(self.sourceColumnNumber)
        s += "\n>"
        return s


class HXLReader(HXLDataProvider):
    """
    Read HXL data from a file
    """

    def __init__(self, source):
        # all internal properties
        self._csv_reader = csv.reader(source)
        self._table_spec = None
        self._source_row_number = -1
        self._row_number = -1
        self._last_header_row = None
        self._raw_data = None
        self._disaggregation_pos = 0

    @property
    def columns(self):
        """
        Return a list of HXLColumn objects.
        """
        self._setup_table_spec()
        return self._table_spec.columns

    def __next__(self):
        """
        Iterable function to return the next row of HXL values.
        Returns a HXLRow, or raises StopIteration exception at end
        """

        # Won't do anything if it already exists
        self._setup_table_spec()

        # Read more raw data unless we're in the middle of generating virtual rows
        # from compact-disaggregated syntax
        if self._disaggregation_pos >= self._table_spec.disaggregationCount or not self._raw_data:
            self._raw_data = self._parse_source_row()
            if (self._raw_data == None):
                return None
            self._disaggregation_pos = 0

        # Next logical row
        self._row_number += 1

        # The row we're going to populate
        row = HXLRow(self._table_spec.columns, rowNumber=self._row_number, sourceRowNumber=self._source_row_number)

        columnNumber = -1
        seenFixed = False

        # Loop through the raw CSV data
        for sourceColumnNumber, content in enumerate(self._raw_data):

            if sourceColumnNumber >= len(self._table_spec.colSpecs):
                # no more hashtags
                break

            # grab the specification
            colSpec = self._table_spec.colSpecs[sourceColumnNumber]

            # We're reading only columns that have HXL tags
            if not colSpec.column.hxlTag:
                continue

            if colSpec.fixedColumn:
                # There's a fixed column involved
                if not seenFixed:
                    columnNumber += 1
                    disaggregatedSourceColumnNumber = self._table_spec.getDisaggregatedSourceColumnNumber(self._disaggregation_pos)
                    row.append(self._table_spec.colSpecs[disaggregatedSourceColumnNumber].fixedValue)
                    columnNumber += 1
                    row.append(self._raw_data[disaggregatedSourceColumnNumber])
                    seenFixed = True
                    self._disaggregation_pos += 1
            else:
                # regular column
                columnNumber += 1
                row.append(self._raw_data[sourceColumnNumber])

        return row

    next = __next__

    def _setup_table_spec(self):
        """
        Go fishing for the HXL hashtag row.
        Returns a _TableSpec on success. Throws an exception on failure.
        """

        # If we already have it, return it
        if (self._table_spec):
            return self._table_spec

        # OK, need to go fishing ...
        try:
            _raw_data = self._parse_source_row()
            while _raw_data is not None:
                _table_spec = self._parse_hashtag_row(_raw_data)
                if (_table_spec != None):
                    self._table_spec = _table_spec
                    return self._table_spec
                else:
                    self._last_header_row = _raw_data
                    _raw_data = self._parse_source_row()
        except StopIteration:
            pass
        raise HXLParseException("HXL hashtag row not found", self._source_row_number)
    
    def _parse_hashtag_row(self, rawDataRow):
        """
        Try parsing the current raw CSV data row as a HXL hashtag row.
        Returns a _TableSpec on success, or None on failure
        """
        _table_spec = _TableSpec()
        seenHeader = 0
        columnNumber = 0
        for sourceColumnNumber, rawString in enumerate(rawDataRow):
            rawString = rawString.strip()
            if rawString:
                colSpec = self._parse_hashtag(columnNumber,sourceColumnNumber, rawString)
                if (colSpec):
                    seenHeader = 1
                    if (colSpec.fixedColumn):
                        colSpec.fixedColumn.headerText = self._pretty_tag(colSpec.fixedColumn.hxlTag)
                        colSpec.column.headerText = self._pretty_tag(colSpec.column.hxlTag)
                        colSpec.fixedValue = self._last_header_row[sourceColumnNumber]
                        columnNumber += 1
                    else:
                        if self._last_header_row and sourceColumnNumber < len(self._last_header_row):
                            colSpec.column.headerText = self._last_header_row[sourceColumnNumber]
                else:
                    return None
            else:
                colSpec = _ColSpec(sourceColumnNumber)
                colSpec.column = HXLColumn(columnNumber, sourceColumnNumber)
            columnNumber += 1
            _table_spec.append(colSpec)

        if seenHeader:
            return _table_spec
        else:
            return None

    def _parse_hashtag(self, columnNumber, sourceColumnNumber, rawString):
        """
        Attempt to parse a full hashtag specification.
        May include compact disaggregated syntax
        Returns a colspec or None
        """

        # Pattern for a single tag
        tagRegexp = '(#[a-zA-Z0-9_]+)(?:\/([a-zA-Z]{2}))?'

        # Pattern for full tag spec (optional second tag following '+')
        fullRegexp = '^\s*' + tagRegexp + '(?:\s*\+\s*' + tagRegexp + ')?$';

        # Try a match
        result = re.match(fullRegexp, rawString)
        if result:
            col1 = HXLColumn(columnNumber, sourceColumnNumber, result.group(1), result.group(2))
            col2 = None

            if result.group(3):
                # There were two tags
                col2 = HXLColumn(columnNumber, sourceColumnNumber, result.group(3), result.group(4))
                colSpec = _ColSpec(sourceColumnNumber, col2, col1)
            else:
                # There was just one tag
                colSpec = _ColSpec(sourceColumnNumber, col1)
            return colSpec
        else:
            return None

    def _parse_source_row(self):
        """
        Parse a row of raw CSV data.
        Returns an array of strings.
        """
        self._source_row_number += 1
        return next(self._csv_reader)

    def _pretty_tag(self, hxlTag):
        """
        Hack a human-readable heading from a HXL tag name.
        """
        hxlTag = re.sub('^#', '', hxlTag)
        hxlTag = re.sub('_(date|deg|id|link|num)$', '', hxlTag)
        hxlTag = re.sub('_', ' ', hxlTag)
        return hxlTag.capitalize()

def readHXL(input, url=None):
    """Load an in-memory HXL dataset."""
    dataset = HXLDataset(url)

    parser = HXLReader(input)
    dataset.columns = parser.columns
    for row in parser:
        dataset.rows.append(row)

    return dataset

def writeHXL(output, source, showHeaders=True):
    """Serialize a HXL dataset to an output stream."""
    for line in genHXL(source, showHeaders):
        output.write(line)

def writeJSON(output, source, showHeaders=True):
    """Serialize a dataset to JSON."""
    for line in genJSON(source, showHeaders):
        output.write(line)

def writeHTML(output, source, showHeaders=True):
    """Serialize a dataset to HTML."""
    for line in genHTML(source, showHeaders):
        output.write(line)

def genHXL(source, showHeaders=True):
    """
    Generate HXL output one row at a time.
    """
    class TextOut:
        """Simple string output source to capture CSV"""
        def __init__(self):
            self.data = ''
        def write(self, s):
            self.data += s
        def get(self):
            data = self.data
            self.data = ''
            return data
    output = TextOut()
    writer = csv.writer(output)
    if showHeaders and source.hasHeaders:
        writer.writerow(source.headers)
        yield output.get()
    writer.writerow(source.tags)
    yield output.get()
    for row in source:
        writer.writerow(row.values)
        yield output.get()

def genJSON(source, showHeaders=True):
    """
    Generate JSON output, one line at a time.
    """
    yield "{\n"
    if showHeaders and source.hasHeaders:
        yield "  \"headers\": " + json.dumps(source.headers) + ",\n"
    yield "  \"tags\": " + json.dumps(source.tags) + ",\n"
    yield "  \"data\": [\n"
    is_first = True
    for row in source:
        if is_first:
            yield "    "
            is_first = False
        else:
            yield ",\n    "
        yield json.dumps(row.values)
    yield "\n  ]\n"
    yield "}\n"

def genHTML(source, showHeaders=True):
    """
    Generate HTML output, one line at a time.
    """
    yield "<table class=\"hxl\">\n"
    yield "  <thead>\n"
    if (showHeaders and source.hasHeaders):
        yield "    <tr class=\"headers\">\n"
        for s in source.headers:
            yield "      <th>" + cgi.escape(s) + "</th>\n"
        yield "    </tr>\n"
    yield "    <tr class=\"tags\">\n"
    for s in source.tags:
        yield "      <th>" + cgi.escape(s) + "</th>\n"
    yield "    </tr>\n"
    yield "  </thead>\n"
    yield "  <tbody>\n"
    type = 'odd'
    for row in source:
        yield "    <tr class=\"data " + type + "\">\n"
        if type == 'odd':
            type = 'even'
        else:
            type = 'odd'
        for s in row:
            yield "      <td>" + cgi.escape(s) + "</td>\n"
        yield "    </tr>\n"
    yield "  </tbody>\n"
    yield "</table>\n"

# end







"""
Select rows from a HXL dataset.
David Megginson
October 2014

Supply a list of simple <hashtag><operator><value> pairs, and return
the rows in the HXL dataset that contain matches for any of them.

License: Public Domain
Documentation: http://hxlstandard.org
"""

import sys
import re
import operator
import argparse
from hxl.model import HXLDataProvider
from hxl.parser import HXLReader, writeHXL

class HXLSelectFilter(HXLDataProvider):
    """
    Composable filter class to select rows from a HXL dataset.

    This is the class supporting the hxlselect command-line utility.

    Because this class is a {@link hxl.model.HXLDataProvider}, you can use
    it as the source to an instance of another filter class to build a
    dynamic, single-threaded processing pipeline.

    Usage:

    <pre>
    source = HXLReader(sys.stdin)
    filter = HXLSelectFilter(source, queries=[('#org', operator.eq, 'OXFAM')])
    writeHXL(sys.stdout, filter)
    </pre>
    """

    def __init__(self, source, queries=[], reverse=False):
        """
        Constructor
        @param source the HXL data source
        @param queries a series for parsed queries
        @param reverse True to reverse the sense of the select
        """
        self.source = source
        self.queries = queries
        self.reverse = reverse

    @property
    def columns(self):
        return self.source.columns

    def __next__(self):
        """
        Return the next row that matches the select.
        """
        row = next(self.source)
        while not self._row_matches_p(row):
            row = next(self.source)
        return row

    next = __next__

    def _try_op(self, op, v1, v2):
        """Try an operator as numeric first, then string"""
        # TODO add dates
        # TODO use knowledge about HXL tags
        try:
            return op(float(v1), float(v2))
        except ValueError:
            return op(v1, v2)

    def _row_matches_p(self, row):
        """Check if a key-value pair appears in a HXL row"""
        for query in self.queries:
            values = row.getAll(query[0])
            if values:
                for value in values:
                    op = query[1]
                    if value and self._try_op(op, value, query[2]):
                        return not self.reverse
        return self.reverse

def operator_re(s, pattern):
    """Regular-expression comparison operator."""
    return re.match(pattern, s)

def operator_nre(s, pattern):
    """Regular-expression negative comparison operator."""
    return not re.match(pattern, s)

# Map of comparison operators
operator_map = {
    '=': operator.eq,
    '!=': operator.ne,
    '<': operator.lt,
    '<=': operator.le,
    '>': operator.gt,
    '>=': operator.ge,
    '~': operator_re,
    '!~': operator_nre
}

#
# Command-line support
#

def parse_query(s):
    """
    Parse a filter expression
    """
    result = re.match('^#?([a-zA-Z][a-zA-Z0-9_]*)([<>]=?|!?=|!?~)(.*)$', s)
    if result:
       filter = list(result.group(1, 2, 3))
       # (re)add hash to start of tag
       filter[0] = '#' + filter[0]
       op = operator_map[filter[1]]
       if op:
           filter[1] = op
           return filter
    print >>stderr, "Bad filter expression: " + s
    exit(2)

def run(args, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr):
    """
    Run hxlselect with command-line arguments.
    @param args A list of arguments, excluding the script name
    @param stdin Standard input for the script
    @param stdout Standard output for the script
    @param stderr Standard error for the script
    """

    # Command-line arguments
    parser = argparse.ArgumentParser(description = 'Filter rows in a HXL dataset.')
    parser.add_argument(
        'infile',
        help='HXL file to read (if omitted, use standard input).',
        nargs='?',
        type=argparse.FileType('r'),
        default=stdin
        )
    parser.add_argument(
        'outfile',
        help='HXL file to write (if omitted, use standard output).',
        nargs='?',
        type=argparse.FileType('w'),
        default=stdout
        )
    parser.add_argument(
        '-q',
        '--query',
        help='query expression for selecting rows (use multiple times for logical OR): <hashtag><op><value>, where <op> is =, !=, <, <=, >, >=, ~, or !~',
        action='append',
        metavar='tag=value, etc.',
        default=[],
        type=parse_query
        )
    parser.add_argument(
        '-r',
        '--reverse',
        help='Show only lines *not* matching criteria',
        action='store_const',
        const=True,
        default=False
        )
    args = parser.parse_args(args)

    # Call the command function
    source = HXLReader(args.infile)
    filter = HXLSelectFilter(source, queries=args.query, reverse=args.reverse)
    writeHXL(args.outfile, filter)

# end



"""
Command function to rename columns in a HXL dataset.
David Megginson
October 2014

License: Public Domain
Documentation: http://hxlstandard.org
"""

import sys
import argparse
import copy
import re
from hxl.model import HXLDataProvider
from hxl.filters import fix_tag
from hxl.parser import HXLReader, writeHXL

class HXLRenameFilter(HXLDataProvider):
    """
    Composable filter class to rename columns in a HXL dataset.

    This is the class supporting the hxlrename command-line utility.

    Because this class is a {@link hxl.model.HXLDataProvider}, you can use
    it as the source to an instance of another filter class to build a
    dynamic, single-threaded processing pipeline.

    Usage:

    <pre>
    source = HXLReader(sys.stdin)
    filter = HXLRenameFilter(source, rename_map={'#x_district': '#adm1'})
    writeHXL(sys.stdout, filter)
    </pre>
    """

    def __init__(self, source, rename_map={}):
        """
        Constructor
        @param source the HXLDataProvider for the data.
        @param rename_map map of tags to rename
        """
        self.source = source
        self.rename_map = rename_map

    @property
    def columns(self):
        """
        Return the renamed columns.
        """
        def rename_tags(column):
            tag = column.hxlTag
            if tag in self.rename_map:
                column = copy.copy(column)
                column.hxlTag = self.rename_map[tag][0]
                if self.rename_map[tag][1]:
                    column.headerText = self.rename_map[tag][1]
            return column
        return map(rename_tags, self.source.columns)

    def __next__(self):
        return next(self.source)

    next = __next__


#
# Command-line support.
#

def run(args, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr):
    """
    Run hxlcut with command-line arguments.
    @param args A list of arguments, excluding the script name
    @param stdin Standard input for the script
    @param stdout Standard output for the script
    @param stderr Standard error for the script
    """

    parser = argparse.ArgumentParser(description = 'Cut columns from a HXL dataset.')
    parser.add_argument(
        'infile',
        help='HXL file to read (if omitted, use standard input).',
        nargs='?',
        type=argparse.FileType('r'),
        default=stdin
        )
    parser.add_argument(
        'outfile',
        help='HXL file to write (if omitted, use standard output).',
        nargs='?',
        type=argparse.FileType('w'),
        default=stdout
        )
    parser.add_argument(
        '-r',
        '--rename',
        help='Rename an old tag to a new one (with an optional new text header).',
        action='append',
        metavar='[#]<original_tag>:[[Text header]#]<new_tag>',
        default=[],
        type=parse_rename
        )
    args = parser.parse_args(args)

    # Call the command function
    source = HXLReader(args.infile)
    filter = HXLRenameFilter(source, dict(args.rename))
    writeHXL(args.outfile, filter)

def parse_rename(s):
    result = re.match('^#?([a-zA-Z][a-zA-Z0-9_]*):(?:(.*)#)?([a-zA-Z][a-zA-Z0-9_]*)', s)
    return ['#' + result.group(1), ['#' + result.group(3), result.group(2)]]

# end



"""
Command function to schema-validate a HXL dataset.
David Megginson
November 2014

License: Public Domain
Documentation: http://hxlstandard.org
"""

import sys
import os
import argparse
from copy import copy
from hxl.model import HXLDataProvider, HXLColumn
from hxl.parser import HXLReader, writeHXL
from hxl.schema import readHXLSchema

class HXLValidateFilter(HXLDataProvider):
    """Composable filter class to validate a HXL dataset against a schema.

    This is the class supporting the hxlvalidate command-line utility.

    Because this class is a {@link hxl.model.HXLDataProvider}, you can use
    it as the source to an instance of another filter class to build a
    dynamic, singled-threaded processing pipeline.

    Usage:

    <pre>
    source = HXLReader(sys.stdin)
    schema = readHXLSchema(readHXL(open('my-schema.csv', 'r')))
    filter = HXLValidateFilter(source, schema)
    writeHXL(sys.stdout, filter)
    </pre>
    """

    def __init__(self, source, schema, show_all=False):
        """
        @param source a HXL data source
        @param schema a HXLSchema object
        @param show_all boolean flag to report all lines (including those without errors).
        """
        self.source = source
        self.schema = schema
        self.show_all = show_all
        self._saved_columns = None

    @property
    def columns(self):
        """
        Add columns for the error reporting.
        """
        if self._saved_columns is None:
            # append error columns
            err_col = HXLColumn(hxlTag='#x_errors', headerText='Error messages')
            tag_col = HXLColumn(hxlTag='#x_tags', headerText='Error tag')
            row_col = HXLColumn(hxlTag='#x_rows', headerText='Error row number (source)')
            col_col = HXLColumn(hxlTag='#x_cols', headerText='Error column number (source)')
            self._saved_columns = self.source.columns + [err_col, tag_col, row_col, col_col]
        return self._saved_columns

    def __next__(self):
        """
        Report rows with error information.
        """
        validation_errors = []
        def callback(error):
            """
            Collect validation errors
            """
            validation_errors.append(error)
        self.schema.callback = callback

        """
        Read rows until we find an error (unless we're printing all rows)
        """
        row = next(self.source)
        while row:
            if not self.schema.validateRow(row) or self.show_all:
                # append error data to row
                error_row = copy(row)
                messages = "\n".join(map(lambda e: e.message, validation_errors))
                tags = "\n".join(map(lambda e: e.rule.hxlTag if e.rule else '', validation_errors))
                rows = "\n".join(map(lambda e: str(e.row.sourceRowNumber) if e.row else '', validation_errors))
                columns = "\n".join(map(lambda e: str(e.column.sourceColumnNumber) if e.column else '', validation_errors))
                error_row.columns = self.columns
                error_row.values = error_row.values + [messages, tags, rows, columns]
                return error_row
            else:
                row = next(self.source)

    next = __next__


def run(args, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr):
    """
    Run hxlvalidate with command-line arguments.
    @param args A list of arguments, excluding the script name
    @param stdin Standard input for the script
    @param stdout Standard output for the script
    @param stderr Standard error for the script
    """

    parser = argparse.ArgumentParser(description = 'Validate a HXL dataset.')
    parser.add_argument(
        'infile',
        help='HXL file to read (if omitted, use standard input).',
        nargs='?',
        type=argparse.FileType('r'),
        default=stdin
        )
    parser.add_argument(
        'outfile',
        help='HXL file to write (if omitted, use standard output).',
        nargs='?',
        type=argparse.FileType('w'),
        default=stdout
        )
    parser.add_argument(
        '-s',
        '--schema',
        help='Schema file for validating the HXL dataset (if omitted, use the default core schema).',
        metavar='schema',
        type=argparse.FileType('r'),
        default=None
        )
    parser.add_argument(
        '-a',
        '--all',
        help='Include all rows in the output, including those without errors',
        action='store_const',
        const=True,
        default=False
        )
    args = parser.parse_args(args)

    source = HXLReader(args.infile)
    if args.schema:
        schema = readHXLSchema(HXLReader(args.schema), baseDir=os.path.dirname(args.schema.name))
    else:
        schema = readHXLSchema()
    filter = HXLValidateFilter(source, schema, args.all)
    writeHXL(args.outfile, filter)

# end



"""
Script to produce GeoJSON from HXL
David Megginson
October 2014

Usage:

  import sys
  from hxl.scripts.hxl2geojson import hxl2geojson

  hxl2geojson(sys.stdin, sys.stdout)

Creates point data only. The script can handle extremely long
datasets, because it does not try to hold the entire dataset in memory
at once.

License: Public Domain
Documentation: http://hxlstandard.org
"""

import sys
import re
import json
from hxl.parser import HXLReader

def hxl2geojson(input, output):
    """
    Convert HXL to GeoJSON
    """

    output.write('{"type": "FeatureCollection",\n')
    output.write(' "features": [\n')

    parser = HXLReader(input)

    pattern = re.compile('^', re.MULTILINE)
    isFirst = True
    for row in parser:
        # Create a feature only when lat/lon are present
        lat = row.get('#lat_deg')
        lon = row.get('#lon_deg')
        if lat and lon:

            try:

                # Construct the basic feature
                feature = {
                    'type': 'Feature',
                    'geometry': {
                        'type': 'Point',
                        'coordinates': [float(lon), float(lat)]
                        },
                    }

                # Add selected properties if present
                properties = {}
                for tag in ['#loctype', '#loc', '#country', '#adm1', '#adm2', '#adm3']:
                    if row.get(tag):
                        properties[tag] = row.get(tag)
                        feature['properties'] = properties
                        
                # Render the feature
                if isFirst:
                    isFirst = False
                else:
                    output.write(',\n')

                output.write(re.sub(pattern, '   ', json.dumps(feature, indent=1)))

            except ValueError:
                sys.stderr.write('Badly formed #lat_deg or #lon_deg: ' + lat + ', ' + lon + '\n')

    output.write('\n ]\n}\n')

# end



"""
Taxonomy support for the Humanitarian Exchange Language (HXL) v1.0
David Megginson
Started January 2015

License: Public Domain
Documentation: http://hxlstandard.org
"""

class HXLTaxonomyException(Exception):

    def __init__(self, message):
        self.message = message

class HXLTaxonomy:

    def __init__(self, terms=None):
        self.terms = terms if terms is not None else {}

    def add(self, term):
        """
        Add a new term
        @param term the new HXLTerm to add
        @return the old term with the same code, if any
        """
        old_term = self.terms.get(term.code)
        self.terms[term.code] = term
        return old_term

    def get(self, code):
        """
        Try getting an existing term
        @param code the code of the term
        @return the term, or None if not found
        """
        return self.terms.get(code)

    def contains(self, code, level=None):
        """
        Check if the taxonomy contains a term
        Will also match the level if provided
        @param code the term's code
        @param level the term's level (or None if not checked)
        @return True if found, False otherwise
        """
        term = self.get(code)
        return True if term and (level is None or int(term.level)==int(level)) else False

    def is_valid(self, callback=None):
        """
        Test whether a taxonomy is referentially-complete and loop-free.
        @param callback a function to call for each error, using the form callback(message, term)
        @return True if the taxonomy is valid; False if not
        """

        def contains_loop(code, ancestors=[]):
            """Test for loops"""
            if not code:
                # successful termination
                return False
            elif code in ancestors:
                # found a loop
                return True
            else:
                term = self.get(code)
                if term:
                    return contains_loop(term.parent_code, ancestors + [code])
                else:
                    return False

        result = True
        for code in self.terms:
            term = self.terms[code]
            if term.parent_code:
                # check for parent
                if not self.terms.get(term.parent_code):
                    if callback is not None:
                        callback("Parent does not exist", term)
                    result = False
                # check for loops
                if contains_loop(term.parent_code):
                    if callback is not None:
                        callback("Loop in ancestors", term)
                    result = False
        return result
        

class HXLTerm:

    __slots__ = ['code', 'parent_code', 'level']

    def __init__(self, code, parent_code=None, level=None):
        self.code = code
        self.parent_code = parent_code
        self.level = level


def readTaxonomy(source):
    """
    Read a taxonomy from a HXL data source.
    Throws a HXLTaxonomyException for duplicate terms.
    @param a HXLDataProvider
    @return a HXLTaxonomy
    """

    taxonomy = HXLTaxonomy()

    for row in source:
        # skip empty rows
        if len(row.values) == 0:
            continue

        # get the code
        code = row.get('#term_id')
        if code:
            parent_code = row.get('#parent_id')
            level = row.get('#level_num')
            term = HXLTerm(code, parent_code, level)
            old_term = taxonomy.add(term)
            if old_term is not None:
                raise HXLTaxonomyException("Duplicate term: " + old_term.code)
        else:
            raise HXLTaxonomyException("Missing #term_id in row " + str(row.sourceRowNumber))

    return taxonomy
                
            



"""
Validation code for the Humanitarian Exchange Language (HXL) v1.0
David Megginson
Started October 2014

License: Public Domain
Documentation: http://hxlstandard.org
"""

import sys
import re
import os
from copy import copy
from email.utils import parseaddr
from .parser import HXLReader
from .taxonomy import readTaxonomy

if sys.version_info[0] > 2:
    from urllib.parse import urlparse
else:
    from urlparse import urlparse


class HXLValidationException(Exception):
    """
    Data structure to hold a HXL validation error.
    """

    def __init__(self, message, rule = None, value = None, row = None, column = None):
        self.message = message
        self.rule = rule
        self.value = value
        self.row = row
        self.column = column

    def __str__(self):
        value = "<no value>"
        if self.value:
            value = str(re.sub('\s+', ' ', self.value))
            value = '"' + value[:10] + (value[10:] and '...') + '"'
        sourceRowNumber = "?"
        if self.row:
            sourceRowNumber = str(self.row.sourceRowNumber + 1)
        sourceColumnNumber = "?"
        if self.column:
            sourceColumnNumber = str(self.column.sourceColumnNumber + 1)
        return "E " + "(" + sourceRowNumber + "," + sourceColumnNumber + ") " + self.rule.hxlTag + " " + value + ": " + self.message

class HXLSchemaRule(object):
    """
    Validation rule for a single HXL hashtag.
    """

    TYPE_TEXT = 1
    TYPE_NUMBER = 2
    TYPE_URL = 3
    TYPE_EMAIL = 4
    TYPE_PHONE = 5

    def __init__(self, hxlTag, minOccur=None, maxOccur=None, dataType=None, minValue=None, maxValue=None,
                 valuePattern=None, valueEnumeration=None, caseSensitive=True, taxonomy=None, taxonomyLevel=None,
                 callback=None):
        self.hxlTag = hxlTag
        self.minOccur = minOccur
        self.maxOccur = maxOccur
        self.dataType = dataType
        self.minValue = minValue
        self.maxValue = maxValue
        self.valuePattern = valuePattern
        self.valueEnumeration = valueEnumeration
        self.caseSensitive = caseSensitive
        self.taxonomy = taxonomy
        self.taxonomyLevel = taxonomyLevel
        self.callback = callback


    def validateRow(self, row):
        """
        Apply the rule to an entire HXLRow
        @param row the HXLRow to validate
        @return True if all matching values in the row are valid
        """

        numberSeen = 0
        result = True

        # Look up only the values that apply to this rule
        values = row.getAll(self.hxlTag)
        if values:
            for columnNumber, value in enumerate(values):
                if not self.validate(value, row, row.columns[columnNumber]):
                    result = False
                if value:
                    numberSeen += 1
        if self.minOccur is not None and numberSeen < self.minOccur:
            result = self._report_error(
                "Expected at least " + str(self.minOccur) + " instance(s) but found " + str(numberSeen),
                row = row
                )
        if self.maxOccur is not None and numberSeen > self.maxOccur:
            result = self._report_error(
                "Expected at most " + str(self.maxOccur) + " instance(s) but found " + str(numberSeen),
                row = row
                )
        return result


    def validate(self, value, row = None, column = None):
        """
        Apply the rule to a single value.
        @param value the value to validate
        @param row (optional) the HXLRow being validated
        @param column (optional) the HXLColumn being validated
        @return True if valid; false otherwise
        """

        if value is None or value == '':
            return True

        result = True
        if not self._test_type(value, row, column):
            result = False
        if not self._test_range(value, row, column):
            result = False
        if not self._test_pattern(value, row, column):
            result = False
        if not self._test_enumeration(value, row, column):
            result = False
        if not self._test_taxonomy(value, row, column):
            result = False

        return result

    def _report_error(self, message, value=None, row=None, column=None):
        """Report an error to the callback."""
        if self.callback != None:
            self.callback(
                HXLValidationException(
                    message=message,
                    rule=self,
                    value = value,
                    row = row,
                    column = column
                    )
                )
        return False

    def _test_type(self, value, row, column):
        """Check the datatype."""
        if self.dataType == self.TYPE_NUMBER:
            try:
                float(value)
                return True
            except ValueError:
                return self._report_error("Expected a number", value, row, column)
        elif self.dataType == self.TYPE_URL:
            pieces = urlparse(value)
            if not (pieces.scheme and pieces.netloc):
                return self._report_error("Expected a URL", value, row, column)
        elif self.dataType == self.TYPE_EMAIL:
            if not re.match('^[^@]+@[^@]+$', value):
                return self._report_error("Expected an email address", value, row, column)
        elif self.dataType == self.TYPE_PHONE:
            if not re.match('^\+?[0-9xX()\s-]{5,}$', value):
                return self._report_error("Expected a phone number", value, row, column)
        
        return True

    def _test_range(self, value, row, column):
        """Test against a numeric range (if specified)."""
        result = True
        try:
            if self.minValue is not None:
                if float(value) < float(self.minValue):
                    result = self._report_error("Value is less than " + str(self.minValue), value, row, column)
            if self.maxValue is not None:
                if float(value) > float(self.maxValue):
                    result = self._report_error("Value is great than " + str(self.maxValue), value, row, column)
        except ValueError:
            result = False
        return result

    def _test_pattern(self, value, row, column):
        """Test against a regular expression pattern (if specified)."""
        if self.valuePattern:
            flags = 0
            if self.caseSensitive:
                flags = re.IGNORECASE
            if not re.match(self.valuePattern, value, flags):
                self._report_error("Failed to match pattern " + str(self.valuePattern), value, row, column)
                return False
        return True

    def _test_enumeration(self, value, row, column):
        """Test against an enumerated set of values (if specified)."""
        if self.valueEnumeration is not None:
            if self.caseSensitive:
                if value not in self.valueEnumeration:
                    return self._report_error("Must be one of " + str(self.valueEnumeration), value, row, column)
            else:
                if value.upper() not in map(lambda item: item.upper(), self.valueEnumeration):
                    return self._report_error("Must be one of " + str(self.valueEnumeration) + " (case-insensitive)", value, row, column)
        return True

    def _test_taxonomy(self, value, row, column):
        """Test against a taxonomy (if specified)."""
        if self.taxonomy is not None:
            if not self.taxonomy.contains(value, self.taxonomyLevel):
                if self.taxonomyLevel is None:
                    return self._report_error("Not in taxonomy", value, row, column)
                else:
                    return self._report_error("Not in taxonomy at level " + str(self.taxonomyLevel), value, row, column)
        return True

    def __str__(self):
        """String representation of a rule (for debugging)"""
        return "<HXL schema rule: " + self.hxlTag + ">"
                
class HXLSchema(object):
    """
    Schema against which to validate a HXL document.
    """

    def __init__(self, rules=[], callback=None):
        self.rules = copy(rules)
        if callback is None:
            self.callback = self.showError
        else:
            self.callback = callback

    def showError(self, error):
        print >> sys.stderr, error
        

    # TODO add support for validating columns against rules, too
    # this is where to mention impossible conditions, or columns
    # without rules

    def validate(self, parser):
        result = True
        for row in parser:
            if not self.validateRow(row):
                result = False
        return result

    def validateRow(self, row):
        result = True
        for rule in self.rules:
            old_callback = rule.callback
            if self.callback:
                rule.callback = self.callback
            if not rule.validateRow(row):
                result = False
            rule.callback = old_callback
        return result

    def __str__(self):
        """String representation of a schema (for debugging)"""
        s = "<HXL schema\n"
        for rule in self.rules:
            s += "  " + str(rule) + "\n"
        s += ">"
        return s

def readHXLSchema(source=None, baseDir=None):
    """
    Load a HXL schema from the provided input stream, or load default schema.
    @param source HXL data source for the scheme (e.g. a HXLReader or filter)
    """

    schema = HXLSchema()

    def parseType(typeString):
        if typeString == 'text':
            return HXLSchemaRule.TYPE_TEXT
        elif typeString == 'number':
            return HXLSchemaRule.TYPE_NUMBER
        elif typeString == 'url':
            return HXLSchemaRule.TYPE_URL
        elif typeString == 'email':
            return HXLSchemaRule.TYPE_EMAIL
        elif typeString == 'phone':
            return HXLSchemaRule.TYPE_PHONE
        else:
            #TODO add warning
            return None

    def toInt(s):
        if s:
            return int(s)
        else:
            return None
        
    def toFloat(s):
        if s:
            return float(s)
        else:
            return None

    def toRegex(s):
        if s:
            return re.compile(s)
        else:
            return None

    def toTaxonomy(s):
        if s:
            if baseDir:
                path = os.path.join(baseDir, s)
            else:
                path = s
            return readTaxonomy(HXLReader(open(path, 'r')))
        else:
            return None

    if source is None:
        path = os.path.join(os.path.dirname(__file__), 'hxl-default-schema.csv');
        input = open(path, 'r')
        source = HXLReader(input)

    for row in source:
        rule = HXLSchemaRule(row.get('#x_tag'))
        rule.minOccur = toInt(row.get('#x_minoccur_num'))
        rule.maxOccur = toInt(row.get('#x_maxoccur_num'))
        rule.dataType = parseType(row.get('#x_datatype'))
        rule.minValue = toFloat(row.get('#x_minvalue_num'))
        rule.maxValue = toFloat(row.get('#x_maxvalue_num'))
        rule.valuePattern = toRegex(row.get('#x_pattern'))
        rule.taxonomy = toTaxonomy(row.get('#x_taxonomy'))
        rule.taxonomyLevel = toInt(row.get('#x_taxonomylevel_num'))
        s = row.get('#x_enumeration')
        if s:
            rule.valueEnumeration = s.split('|')
        rule.caseSensitive = toInt(row.get('#x_casesensitive'))
        schema.rules.append(rule)

    return schema

# end



def fix_tag(t):
    '''trim whitespace and add # if needed'''
    t = t.strip()
    if not t.startswith('#'):
        t = '#' + t
    return t

def parse_tags(s):
    '''Parse tags out from a comma-separated list'''
    return list(map(fix_tag, s.split(',')))

def find_column(tag, columns):
    '''Find the first column in a list with tag'''
    for column in columns:
        if column.hxlTag == tag:
            return column
    return None



"""
Sort a HXL dataset.
David Megginson
January 2015

The HXLSortFilter class will use numeric sorting for hashtags ending
in _num or _deg, and date-normalised sorting for hashtags ending in
_date.

Warning: this filter reads the entire source dataset into memory
before sorting it.

License: Public Domain
Documentation: http://hxlstandard.org
"""

import sys
import argparse
import dateutil.parser
from hxl.model import HXLDataProvider
from hxl.parser import HXLReader, writeHXL
from hxl.filters import parse_tags

class HXLSortFilter(HXLDataProvider):
    """
    Composable filter class to sort a HXL dataset.

    This is the class supporting the hxlsort command-line utility.

    Because this class is a {@link hxl.model.HXLDataProvider}, you can use
    it as the source to an instance of another filter class to build a
    dynamic, single-threaded processing pipeline.

    Usage:

    <pre>
    source = HXLReader(sys.stdin)
    filter = HXLSortFilter(source, tags=['#sector', '#org', '#adm1'])
    writeHXL(sys.stdout, filter)
    </pre>
    """

    def __init__(self, source, tags=[], reverse=False):
        """
        @param source a HXL data source
        @param tags tags for sorting
        @param reverse True to reverse the sort order
        """
        self.source = source
        self.sort_tags = tags
        self.reverse = reverse
        self._iter = None

    @property
    def columns(self):
        """
        Return the same columns as the source.
        """
        return self.source.columns

    def __next__(self):
        """
        Sort the dataset first, then return it row by row.
        """

        # Closures
        def make_key(row):
            """Closure: use the requested tags as keys (if provided). """

            def get_value(tag):
                """Closure: extract a sort key"""
                raw_value = row.get(tag)
                if tag.endswith('_num') or tag.endswith('_deg'):
                    # left-pad numbers with zeros
                    try:
                        raw_value = str(float(raw_value)).zfill(15)
                    except ValueError:
                        pass
                elif tag.endswith('_date'):
                    # normalise dates for sorting
                    raw_value = dateutil.parser.parse(raw_value).strftime('%Y-%m-%d')
                return raw_value.upper()

            if self.sort_tags:
                return list(map(get_value, self.sort_tags))
            else:
                return row.values

        # Main method
        if self._iter is None:
            self._iter = iter(sorted(self.source, key=make_key, reverse=self.reverse))
        return next(self._iter)

    next = __next__

            
#
# Command-line support
#

def run(args, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr):
    """
    Run hxlcut with command-line arguments.
    @param args A list of arguments, excluding the script name
    @param stdin Standard input for the script
    @param stdout Standard output for the script
    @param stderr Standard error for the script
    """

    parser = argparse.ArgumentParser(description = 'Sort a HXL dataset.')
    parser.add_argument(
        'infile',
        help='HXL file to read (if omitted, use standard input).',
        nargs='?',
        type=argparse.FileType('r'),
        default=stdin
        )
    parser.add_argument(
        'outfile',
        help='HXL file to write (if omitted, use standard output).',
        nargs='?',
        type=argparse.FileType('w'),
        default=stdout
        )
    parser.add_argument(
        '-t',
        '--tags',
        help='Comma-separated list of tags to for columns to use as sort keys.',
        metavar='tag,tag...',
        type=parse_tags
        )
    parser.add_argument(
        '-r',
        '--reverse',
        help='Flag to reverse sort order.',
        action='store_const',
        const=True,
        default=False
        )
    args = parser.parse_args(args)

    # Call the command function
    source = HXLReader(args.infile)
    filter = HXLSortFilter(source, args.tags, args.reverse)
    writeHXL(args.outfile, filter)

# end



"""
Data model for the Humanitarian Exchange Language (HXL) v1.0
David Megginson
Started October 2014

License: Public Domain
Documentation: http://hxlstandard.org
"""

import abc
from copy import copy

class HXLDataProvider(object):
    """
    Abstract base class for a HXL data source.

    Any source of parsed HXL data inherits from this class: that
    includes HXLDataset, HXLReader, and the various filters in the
    hxl.filters package.  The contract of a HXLDataProvider is that it will
    provide a columns property and a next() method to read through the
    rows.

    The child class must implement the columns() method as a property
    and the next() method to iterate through rows of data.
    """

    __metaclass__ = abc.ABCMeta

    def __iter__(self):
        """
        The object itself is an iterator.
        """
        return self

    @property
    @abc.abstractmethod
    def columns(self):
        """
        Get the column definitions for the dataset.
        @return a list of HXLColumn objects.
        """
        return

    @abc.abstractmethod
    def __next__(self):
        """
        Iterable function to return the next row of HXL values.
        @return an iterable HXLRow
        @exception StopIteration exception at end of the rows.
        """
        return

    @property
    def headers(self):
        """
        Return a list of header strings (for a spreadsheet row).
        """
        return list(map(lambda column: column.headerText, self.columns))

    @property
    def tags(self):
        """
        Return a list of tags.
        """
        return list(map(lambda column: column.hxlTag, self.columns))

    @property
    def displayTags(self):
        """
        Return a list of display tags.
        """
        return list(map(lambda column: column.displayTag, self.columns))

    @property
    def hasHeaders(self):
        """
        Report whether any non-empty header strings exist.
        """
        for column in self.columns:
            if column.headerText:
                return True
        return False


class HXLDataset(HXLDataProvider):
    """
    In-memory HXL dataset.
    """

    def __init__(self, url=None, columns=[], rows=[]):
        """
        Initialise a dataset.
        @param url The dataset's URL (default: None).
        """
        self.url = url
        self.columns = copy(columns)
        self.rows = copy(rows)

    def __str__(self):
        """
        Create a string representation of a dataset for debugging.
        @return A debugging string.
        """
        if self.url:
            return '<HXLDataset ' + self.url + '>'
        else:
            return '<HXLDataset>'


class HXLColumn(object):
    """
    The definition of a logical column in the HXL data.
    """ 

    # To tighten debugging (may reconsider later -- not really a question of memory efficiency here)
    __slots__ = ['columnNumber', 'sourceColumnNumber', 'hxlTag', 'languageCode', 'headerText']

    def __init__(self, columnNumber=None, sourceColumnNumber=None, hxlTag=None, languageCode=None, headerText=None):
        """
        Initialise a column definition.
        @param columnNumber the logical column number (default: None)
        @param sourceColumnNumber the raw column number in the source dataset (default: None)
        @param hxlTag the HXL hashtag for the column (default: None)
        @param languageCode the ISO 639- language code for the column, e.g. "es" (default: None)
        @param headerText the original plaintext header for the column (default: None)
        """
        self.columnNumber = columnNumber
        self.sourceColumnNumber = sourceColumnNumber
        self.hxlTag = hxlTag
        self.languageCode = languageCode
        self.headerText = headerText

    @property
    def displayTag(self):
        """
        Generate a display version of the column hashtag
        @return the reassembled HXL hashtag string, including language code
        """
        if (self.hxlTag):
            if (self.languageCode):
                return self.hxlTag + '/' + self.languageCode
            else:
                return self.hxlTag
        else:
            return None

    def __str__(self):
        """
        Create a string representation of a column header for debugging.
        """
        tag = self.displayTag
        if tag:
            return '<HXLColumn ' + str(tag) + '>'
        else:
            return '<HXLColumn>'


class HXLRow(object):
    """
    An iterable row of values in a HXL dataset.
    """

    # Predefine the slots for efficiency (may reconsider later)
    __slots__ = ['columns', 'values', 'rowNumber', 'sourceRowNumber']

    def __init__(self, columns, rowNumber=None, sourceRowNumber=None, values=[]):
        """
        Set up a new row.
        @param columns The column definitions (array of HXLColumn objects).
        @param rowNumber The logical row number in the input dataset (default: None)
        @param sourceRowNumber The original row number in the raw source dataset (default: None)
        """
        self.columns = columns
        self.values = copy(values)
        self.rowNumber = rowNumber
        self.sourceRowNumber = sourceRowNumber

    def append(self, value):
        """
        Append a value to the row.
        @param value The new value to append.
        @return The new value
        """
        self.values.append(value)
        return value

    def get(self, tag, index=0, default=None):
        """
        Get a single value for a tag in a row.
        @param tag The HXL tag for the value.
        @param index The zero-based index if there are multiple values for the tag (default: 0)
        @param default The default value if not found (default: None)
        @return The value found, or the default value provided.
        """
        for i, column in enumerate(self.columns):
            if i >= len(self.values):
                break
            if column.hxlTag == tag:
                if index == 0:
                    return self.values[i]
                else:
                    index = index - 1
        return default

    def getAll(self, tag):
        """
        Get all values for a specific tag in a row
        @param tag The HXL tag for the value(s).
        @return An array of values for the HXL hashtag.
        """
        result = []
        for i, column in enumerate(self.columns):
            if i >= len(self.values):
                break
            if column.hxlTag == tag:
                result.append(self.values[i])
        return result

    def map(self, function):
        """
        Map a function over a row and return the result.
        
        The function returns a new list constructed from the return
        values of the mapping function, which must take two arguments
        (the value, and the HXLColumn object).
        @param function The mapping function.
        @return A new array of values, after the mapping
        """
        values = []
        for index, value in enumerate(self.values):
            values.append(function(value, self.columns[index]))
        return values

    def __getitem__(self, index):
        """
        Array-access method to make this class iterable.
        @param index The zero-based index of a value to look up.
        @return The value if it exists.
        @exception IndexError if the index is out of range.
        """
        return self.values[index]

    def __str__(self):
        """
        Create a string representation of a row for debugging.
        """
        s = '<HXLRow';
        s += "\n  rowNumber: " + str(self.rowNumber)
        s += "\n  sourceRowNumber: " + str(self.sourceRowNumber)
        for columnNumber, value in enumerate(self.values):
            s += "\n  " + str(self.columns[columnNumber]) + "=" + str(value)
        s += "\n>"
        return s

# end



"""
Cut columns from a HXL dataset.
David Megginson
October 2014

Can use a whitelist of HXL tags, a blacklist, or both.

License: Public Domain
Documentation: http://hxlstandard.org
"""

import sys
import argparse
from hxl.model import HXLDataProvider, HXLRow
from hxl.parser import HXLReader, writeHXL
from hxl.filters import parse_tags

class HXLCutFilter(HXLDataProvider):
    """
    Composable filter class to cut columns from a HXL dataset.

    This is the class supporting the hxlcut command-line utility.

    Because this class is a {@link hxl.model.HXLDataProvider}, you can use
    it as the source to an instance of another filter class to build a
    dynamic, single-threaded processing pipeline.

    Usage:

    <pre>
    source = HXLReader(sys.stdin)
    filter = HXLCutFilter(source, include_tags=['#sector', '#org', '#adm1'])
    writeHXL(sys.stdout, filter)
    </pre>
    """

    def __init__(self, source, include_tags=[], exclude_tags=[]):
        """
        @param source a HXL data source
        @param include_tags a whitelist list of hashtags to include
        @param exclude_tags a blacklist of hashtags to exclude
        """
        self.source = source
        self.include_tags = include_tags
        self.exclude_tags = exclude_tags
        self.columns_out = None

    @property
    def columns(self):
        """
        Filter out the columns that should be removed.
        """
        if self.columns_out is None:
            self.columns_out = []
            for column in self.source.columns:
                if self._test_column(column):
                    self.columns_out.append(column)
        return self.columns_out

    def __next__(self):
        """
        Return the next row, with appropriate columns filtered out.
        """
        row = next(self.source)
        values_out = []
        for pos, value in enumerate(row):
            if self._test_column(row.columns[pos]):
                values_out.append(value)
        row_out = HXLRow(self.columns)
        row_out.values = values_out
        return row_out

    next = __next__

    def _test_column(self, column):
        """
        Test whether a column should be included in the output.
        If there is a whitelist, it must be in the whitelist; if there is a blacklist, it must not be in the blacklist.
        """
        return ((not self.include_tags) or (column.hxlTag in self.include_tags)) and ((not self.exclude_tags) or (column.hxlTag not in self.exclude_tags))

#
# Command-line support
#

def run(args, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr):
    """
    Run hxlcut with command-line arguments.
    @param args A list of arguments, excluding the script name
    @param stdin Standard input for the script
    @param stdout Standard output for the script
    @param stderr Standard error for the script
    """

    parser = argparse.ArgumentParser(description = 'Cut columns from a HXL dataset.')
    parser.add_argument(
        'infile',
        help='HXL file to read (if omitted, use standard input).',
        nargs='?',
        type=argparse.FileType('r'),
        default=stdin
        )
    parser.add_argument(
        'outfile',
        help='HXL file to write (if omitted, use standard output).',
        nargs='?',
        type=argparse.FileType('w'),
        default=stdout
        )
    parser.add_argument(
        '-i',
        '--include',
        help='Comma-separated list of column tags to include',
        metavar='tag,tag...',
        type=parse_tags
        )
    parser.add_argument(
        '-x',
        '--exclude',
        help='Comma-separated list of column tags to exclude',
        metavar='tag,tag...',
        type=parse_tags
        )
    args = parser.parse_args(args)

    # Call the command function
    source = HXLReader(args.infile)
    filter = HXLCutFilter(source, args.include, args.exclude)
    writeHXL(args.outfile, filter)

# end



"""
Command function to merge multiple HXL datasets.
David Megginson
November 2014

Can use a whitelist of HXL tags, a blacklist, or both.

License: Public Domain
Documentation: http://hxlstandard.org
"""

import sys
import argparse
from copy import copy
from hxl.model import HXLDataProvider, HXLColumn
from hxl.parser import HXLReader, writeHXL
from hxl.filters import parse_tags, find_column

class HXLMergeFilter(HXLDataProvider):
    """
    Composable filter class to merge values from two HXL datasets.

    This is the class supporting the hxlmerge command-line utility.

    Warning: this filter may store a large amount of data in memory, depending on the merge.

    Because this class is a {@link hxl.model.HXLDataProvider}, you can use
    it as the source to an instance of another filter class to build a
    dynamic, single-threaded processing pipeline.

    Usage:

    <pre>
    source = HXLReader(sys.stdin)
    merge_source = HXLReader(open('file-to-merge.csv', 'r'))
    filter = HXLMergeFilter(source, merge_source=merge_source, keys=['adm1_id'], tags=['adm1'])
    writeHXL(sys.stdout, filter)
    </pre>
    """

    def __init__(self, source, merge_source, keys, tags, before=False):
        """
        Constructor.
        @param source the HXL data source.
        @param merge_source a second HXL data source to merge into the first.
        @param keys the shared key hashtags to use for the merge
        @param tags the tags to include from the second dataset
        @param before if True, add new columns before existing ones
        """
        self.source = source
        self.merge_source = merge_source
        self.keys = keys
        self.merge_tags = tags
        self.before = before
        self.saved_columns = None
        self.merge_map = None
        self.empty_result = [''] * len(tags)

    @property
    def columns(self):
        """
        @return column definitions for the merged dataset
        """
        if self.saved_columns is None:
            new_columns = []
            for tag in self.merge_tags:
                column = find_column(tag, self.merge_source.columns)
                if column:
                    headerText = column.headerText
                else:
                    headerText = None
                new_columns.append(HXLColumn(hxlTag=tag, headerText=headerText))
            if self.before:
                self.saved_columns =  new_columns + self.source.columns
            else:
                self.saved_columns = self.source.columns + new_columns
        return self.saved_columns

    def __next__(self):
        """
        @return the next merged row of data
        """
        if self.merge_map is None:
            self.merge_map = self._read_merge()
        row = copy(next(self.source))
        merge_values = self.merge_map.get(self._make_key(row))
        if not merge_values:
            merge_values = self.empty_result
        if self.before:
            row.values = merge_values + row.values
        else:
            row.values = row.values + merge_values
        return row

    next = __next__

    def _make_key(self, row):
        """
        Make a tuple key for a row.
        """
        values = []
        for key in self.keys:
            values.append(row.get(key))
        return tuple(values)

    def _read_merge(self):
        """
        Read the second (merging) dataset into memory.
        Stores only the values necessary for the merge.
        @return a map of merge values
        """
        merge_map = {}
        for row in self.merge_source:
            values = []
            for tag in self.merge_tags:
                values.append(row.get(tag))
            merge_map[self._make_key(row)] = values
        return merge_map

#
# Command-line support
#

def run(args, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr):
    """
    Run hxlmerge with command-line arguments.
    @param args A list of arguments, excluding the script name
    @param stdin Standard input for the script
    @param stdout Standard output for the script
    @param stderr Standard error for the script
    """

    parser = argparse.ArgumentParser(description = 'Merge part of one HXL dataset into another.')
    parser.add_argument(
        'infile',
        help='HXL file to read (if omitted, use standard input).',
        nargs='?',
        type=argparse.FileType('r'),
        default=stdin
        )
    parser.add_argument(
        'outfile',
        help='HXL file to write (if omitted, use standard output).',
        nargs='?',
        type=argparse.FileType('w'),
        default=stdout
        )
    parser.add_argument(
        '-m',
        '--merge',
        help='HXL file to write (if omitted, use standard output).',
        metavar='filename',
        required=True,
        type=argparse.FileType('r')
        )
    parser.add_argument(
        '-k',
        '--keys',
        help='HXL tag(s) to use as a shared key.',
        metavar='tag,tag...',
        required=True,
        type=parse_tags
        )
    parser.add_argument(
        '-t',
        '--tags',
        help='Comma-separated list of column tags to include from the merge dataset.',
        metavar='tag,tag...',
        required=True,
        type=parse_tags
        )
    parser.add_argument(
        '-b',
        '--before',
        help='Add new columns before existing ones rather than after them.',
        action='store_const',
        const=True,
        default=False
    )
    args = parser.parse_args(args)

    source = HXLReader(args.infile)
    filter = HXLMergeFilter(source, merge_source=HXLReader(args.merge), keys=args.keys, tags=args.tags, before=args.before)
    writeHXL(args.outfile, filter)

# end



#!/usr/bin/python

from setuptools import setup
dependency_links=[
    'git+https://github.com/Toblerity/Shapely.git@maint#egg=Shapely',
]

setup(name='libhxl',
      version='0.12',
      description='Python support for the Humanitarian Exchange Language (HXL).',
      author='David Megginson',
      author_email='contact@megginson.com',
      url='http://hxlproject.org',
      install_requires=['shapely', 'python-dateutil'],
      dependency_links=dependency_links,
      packages=['hxl', 'hxl.filters'],
      package_data={'': ['*.csv']},
      include_package_data=True,
      test_suite='tests',
      scripts=[
        'scripts/hxl2geojson',
        'scripts/hxladd',
        'scripts/hxlbounds',
        'scripts/hxlcount',
        'scripts/hxlcut',
        'scripts/hxlmerge',
        'scripts/hxlnorm',
        'scripts/hxlrename',
        'scripts/hxlselect',
        'scripts/hxlsort',
        'scripts/hxlvalidate'
        ]
      )



"""
Add constant values to a HXL dataset.
David Megginson
January 2015

License: Public Domain
Documentation: http://hxlstandard.org
"""

import sys
import argparse
import re
from copy import copy
from hxl.model import HXLDataProvider, HXLColumn
from hxl.parser import HXLReader, writeHXL
from hxl.filters import parse_tags

class HXLAddFilter(HXLDataProvider):
    """
    Composable filter class to add constant values to every row of a HXL dataset.

    This is the class supporting the hxladd command-line utility.

    Because this class is a {@link hxl.model.HXLDataProvider}, you can use
    it as the source to an instance of another filter class to build a
    dynamic, single-threaded processing pipeline.

    Usage:

    <pre>
    source = HXLReader(sys.stdin)
    filter = HXLAddFilter(source, values={'#report_date': '2015-03-01', '#country': 'Kenya'})
    writeHXL(sys.stdout, filter)
    </pre>
    """

    def __init__(self, source, values, before=False):
        """
        @param source a HXL data source
        @param values a dictionary of tags and constant values
        @param before True to add new columns before existing ones
        """
        self.source = source
        self.values = values
        self.before = before
        self._columns_out = None
        self._const_values = [self.values[tag][0] for tag in self.values]

    @property
    def columns(self):
        """
        Add the constant columns to the end.
        """
        if self._columns_out is None:
            new_columns = []
            for tag in self.values:
                new_columns.append(HXLColumn(hxlTag=tag, headerText=self.values[tag][1]))
            if self.before:
                self._columns_out = new_columns + self.source.columns
            else:
                self._columns_out = self.source.columns + new_columns
        return self._columns_out

    def __next__(self):
        """
        Return the next row, with constant values added.
        """
        row = copy(next(self.source))
        if self.before:
            row.values = self._const_values + row.values
        else:
            row.values = row.values + self._const_values
        return row

    next = __next__


#
# Command-line support
#

def parse_value(s):
    """Parse a tag=value statement."""
    result = re.match('^(?:(.*)#)?([a-zA-Z][a-zA-Z0-9_]*)=(.*)$', s)
    if result:
        items = [result.group(2), result.group(3, 1)]
        # make sure the tag starts with '#'
        if not items[0].startswith('#'):
            items[0] = '#' + items[0]
        return items
    else:
        raise Exception("Bad value expression: " + s)

def run(args, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr):
    """
    Run hxladd with command-line arguments.
    @param args A list of arguments, excluding the script name
    @param stdin Standard input for the script
    @param stdout Standard output for the script
    @param stderr Standard error for the script
    """

    parser = argparse.ArgumentParser(description = 'Add new columns with constant values to a HXL dataset.')
    parser.add_argument(
        'infile',
        help='HXL file to read (if omitted, use standard input).',
        nargs='?',
        type=argparse.FileType('r'),
        default=stdin
        )
    parser.add_argument(
        'outfile',
        help='HXL file to write (if omitted, use standard output).',
        nargs='?',
        type=argparse.FileType('w'),
        default=stdout
        )
    parser.add_argument(
        '-v',
        '--value',
        help='Constant value to add to each row',
        metavar='[[Text header]#]<tag>=<value>',
        action='append',
        required=True,
        type=parse_value
        )
    parser.add_argument(
        '-b',
        '--before',
        help='Add new columns before existing ones rather than after them.',
        action='store_const',
        const=True,
        default=False
    )
        
    args = parser.parse_args(args)

    # Call the command function
    source = HXLReader(args.infile)
    filter = HXLAddFilter(source, values=dict(args.value), before=args.before)
    writeHXL(args.outfile, filter)

# end



"""
Script to perform bounds checking on a HXL dataset.
David Megginson
October 2014

Check that all #lat_deg and #lon_deg values are contained within the
#GeoJSON featureset provided.

Prerequisites:
- Python Shapely library must be installed: https://pypi.python.org/pypi/Shapely
- libgeos_c must be installed: 

In Ubuntu Linux, the following commands will do the trick:

  sudo apt-get install libgeos-c1
  sudo pip install Shapely

Command-line usage:

  python -m hxl.scripts.hxlbounds -b BOUNDS.json DATA.csv > report.txt

(Use -h option to get full usage.)

Program usage:

  import sys
  from hxl.scripts.hxlbonds import hxlbounds

  data = json.load(open('LBR.json', 'r'))
  bounds = []
  for d in data['features']:
      bounds.append(shape(d['geometry']))
  hxlbounds(sys.stdin, sys.stdout, bounds)

License: Public Domain
Documentation: http://hxlstandard.org
"""

import sys
import argparse
import json
from shapely.geometry import Point, shape
from hxl.parser import HXLReader
from . import parse_tags

def hxlbounds(input, output, bounds, tags=[]):
    """
    Check that all points in a HXL dataset fall without a set of bounds.
    """

    def error(row, message):
        """Report a bounds error."""
        lat = row.get('#lat_deg')
        lon = row.get('#lon_deg')
        context = [
            '#lat_deg' + '=' + lat,
            '#lon_deg' + '=' + lon
            ]
        if (tags):
            for tag in tags:
                value = row.get(tag)
                if value:
                    context.append(tag + '=' + value)
        report = str(message) + ' (row ' + str(row.sourceRowNumber) + ') ' + str(context)
        print >>output, report

    reader = HXLReader(input)
    for row in reader:
        lat = row.get('#lat_deg')
        lon = row.get('#lon_deg')
        if lat and lon:
            try:
                seen_shape = False
                for s in bounds:
                    if s.contains(Point(float(lon), float(lat))):
                        seen_shape = True
                        break;
                if not seen_shape:
                    error(row, 'out of bounds')
            except ValueError:
                error(row, 'malformed lat/lon')
        elif lat or lon:
            error(row, '#lat_deg or #lon_deg missing')
        # TODO user option to report no lat *and* no lon
                    

# end



"""
Command function to normalise a HXL dataset.
David Megginson
October 2014

License: Public Domain
Documentation: http://hxlstandard.org
"""

import sys
import re
import dateutil.parser
import argparse
from copy import copy
from hxl.model import HXLDataProvider
from hxl.parser import HXLReader, writeHXL
from hxl.filters import parse_tags

class HXLNormFilter(HXLDataProvider):

    def __init__(self, source, whitespace=False, upper=[], lower=[], date=[], number=[]):
        self.source = source
        self.whitespace = whitespace
        self.upper = upper
        self.lower = lower
        self.date = date
        self.number = number

    @property
    def columns(self):
        return self.source.columns

    def __next__(self):

        def normalise(value, column):
            """
            Normalise a single HXL value.
            """

            # Whitespace (-w or -W)
            if self.whitespace:
                if (self.whitespace is True) or (column.hxlTag in self.whitespace):
                    value = re.sub('^\s+', '', value)
                    value = re.sub('\s+$', '', value)
                    value = re.sub('\s+', ' ', value)

            # Uppercase (-u)
            if self.upper and column.hxlTag in self.upper:
                if sys.version_info[0] > 2:
                    value = value.upper()
                else:
                    value = value.decode('utf8').upper().encode('utf8')

            # Lowercase (-l)
            if self.lower and column.hxlTag in self.lower:
                if sys.version_info[0] > 2:
                    value = value.lower()
                else:
                    value = value.decode('utf8').lower().encode('utf8')

            # Date (-d or -D)
            if self.date and value:
                if (self.date is True and column.hxlTag.endswith('_date')) or (self.date is not True and column.hxlTag in self.date):
                    value = dateutil.parser.parse(value).strftime('%Y-%m-%d')

            # Number (-n or -N)
            if self.number and re.match('\d', value):
                if (self.number is True and column.hxlTag.endswith('_num')) or (self.number is not True and column.hxlTag in self.number):
                    value = re.sub('[^\d.]', '', value)
                    value = re.sub('^0+', '', value)
                    value = re.sub('(\..*)0+$', '\g<1>', value)
                    value = re.sub('\.$', '', value)

            return value

        row = copy(next(self.source))
        row.values = row.map(normalise)
        return row

    next = __next__


def run(args, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr):
    """
    Run hxlnorm with command-line arguments.
    @param args A list of arguments, excluding the script name
    @param stdin Standard input for the script
    @param stdout Standard output for the script
    @param stderr Standard error for the script
    """

    # Command-line arguments
    parser = argparse.ArgumentParser(description = 'Normalize a HXL file.')
    parser.add_argument(
        'infile',
        help='HXL file to read (if omitted, use standard input).',
        nargs='?',
        type=argparse.FileType('r'),
        default=stdin
        )
    parser.add_argument(
        'outfile',
        help='HXL file to write (if omitted, use standard output).',
        nargs='?',
        type=argparse.FileType('w'),
        default=stdout
        )
    parser.add_argument(
        '-W',
        '--whitespace-all',
        help='Normalise whitespace in all columns',
        action='store_const',
        const=True,
        default=False
        )
    parser.add_argument(
        '-w',
        '--whitespace',
        help='Comma-separated list of tags for normalised whitespace.',
        metavar='tag,tag...',
        type=parse_tags
        )
    parser.add_argument(
        '-u',
        '--upper',
        help='Comma-separated list of tags to convert to uppercase.',
        metavar='tag,tag...',
        type=parse_tags
        )
    parser.add_argument(
        '-l',
        '--lower',
        help='Comma-separated list of tags to convert to lowercase.',
        metavar='tag,tag...',
        type=parse_tags
        )
    parser.add_argument(
        '-D',
        '--date-all',
        help='Normalise all dates.',
        action='store_const',
        const=True,
        default=False
        )
    parser.add_argument(
        '-d',
        '--date',
        help='Comma-separated list of tags for date normalisation.',
        metavar='tag,tag...',
        type=parse_tags
        )
    parser.add_argument(
        '-N',
        '--number-all',
        help='Normalise all numbers.',
        action='store_const',
        const=True,
        default=False
        )
    parser.add_argument(
        '-n',
        '--number',
        help='Comma-separated list of tags for number normalisation.',
        metavar='tag,tag...',
        type=parse_tags
        )
    parser.add_argument(
        '-H',
        '--headers',
        help='Preserve text header row above HXL hashtags',
        action='store_const',
        const=True,
        default=False
        )
    args = parser.parse_args(args)

    # Call the command function
    if args.whitespace_all:
        whitespace_arg = True
    else:
        whitespace_arg = args.whitespace

    if args.date_all:
        date_arg = True
    else:
        date_arg = args.date

    if args.number_all:
        number_arg = True
    else:
        number_arg = args.number

    source = HXLReader(args.infile)
    filter = HXLNormFilter(source, whitespace=whitespace_arg, upper=args.upper, lower=args.lower, date=date_arg, number=number_arg)
    writeHXL(args.outfile, filter, args.headers)

# end
